from __future__ import annotations

class WBTextMessage:
  def __init__(self, req: dict) -> WBTextMessage:
    self.req = req['message']

  @property
  def content(self):
    return self.req['text']

  text = content

  @property
  def id(self):
    return self.req['id']


class WBPostback:
  def __init__(self, req: dict) -> WBPostback:
    self.req = req['postback']
    self.TYPE = 'postback'

  @property
  def data(self):
    return self.req['data']

  @property
  def datetime(self):
    res = self.req.get('params', {"datetime": None}).get('datetime')
    if res:
      self.TYPE = 'datetime'
    return res

  @property
  def rich_menu(self):
    if self.TYPE != 'postback':
      rm = self.req.get('params', {
        # NOT FOUND value (default value)
        "newRichMenuAliasId": None,
        "status": None
      })
      class RichMenuSwitchEvent:
        status: str = rm['status']
        new_rich_menu_alias_id = new_rich_menu = rm['newRichMenuAliasId']
  
      res = RichMenuSwitchEvent()
      if res.new_rich_menu_alias_id:
        self.TYPE = 'rich_menu_switch'
      return res
    return None

class WBSticker:
    def __init__(self, req: dict) -> WBSticker:
        self.req = req['message']

    @property
    def sticker_id(self):
        return self.req['stickerId']

    @property
    def package_id(self):
        return self.req['packageId']

    @property
    def resource_type(self):
        return self.req['stickerResourceType'].lower()

    @property
    def keywords(self):
        return self.req['keywords']

# bro think he smart :skull:


from urllib.parse import quote
from .database import *
from . import (
  commands as commands,
  rule
)

class SocialPlugins:
  """
  The LINE Social Plugins. (extension)
  """
  @staticmethod
  def share(url: str) -> str:
    """
    Returns a LINE Share Link. This is useful when you want a custom button.
    """
    return "https://social-plugins.line.me/lineit/share?url=" + str(quote(url))


class _depends:
  def __bool__(self):
    return False

  def __repr__(self):
    return "It Depends."

  def __eq__(self):
    return False

  def __hash__(self):
    return 0

  def __getitem__(self, k):
    return k


Depends = _depends()


from __future__ import annotations

from typing import Literal
import time

from ..tmp import Tmp

# const
VALID_RULES = {
    # tuple: low usage.
    "cooldown": ('seconds',), 
    "except": ('users',), 
    "for": ('users',),
    "based.custom": (),
    "usage_limit": ('times',)
}

class _dfr:
    def emit(self, *args, **kwargs) -> bool:
        return True

# export
DEFAULT_RULE = _dfr()

# export
class CommandRule:
    """
    Represents a command rule. The specification depends on the rule type.

    ## cooldown
    Represents a command cooldown. The user should wait until the cooldown (in seconds) ends, or else a specific command will not work.

    > Arg Requires: `seconds: int`

    ```py
    CommandRule(
      rule='cooldown',
      seconds=10
    )
    ```

    ## except
    Run the command except for someone. The specified users will not be able to use the command. An array of user ids (string) should be passed.

    > Arg Requires: `users: list[str]`

    ```py
    CommandRule(
      rule='except',
      users=[
        'user id 1',
        'user id 2'
      ]
    )
    ```

    ## for
    Run the command for some people. The users that are not on the list will not be able to use this command. An array of user ids (string) should be passed.

    > Arg Requires: `users: list[str]`

    ```py
    CommandRule(
      rule='for',
      users=[
        'user id 1',
        'user id 2'
      ]
    )
    ```

    ## based.custom
    Based on your custom rule. Linelib will not touch anything from your function.

    > Arg Requires: No additional arguments required for this condition.

    ```py
    class MyRule(CommandRule):
      def handler(self, ctx):
        return True # true or false

    MyRule(rule='based.custom') # now it's a valid rule.
    ```

    ## usage_limit
    Limit a command based on the usage of it. For instance, using a usage limitation for each user, everyone will only be able to use it for a specified time.

    > Arg Requires: `times: int`

    ```py
    # the command could only be used once
    
    CommandRule(
      rule='usage_limit',
      times=1
    )
    ```
    """
    rule: str

    def __init__(self, *, rule: Literal["cooldown", "except", "for", "based.custom", "usage_limit"], **variations):
        if not rule in VALID_RULES:
            raise ValueError(f"\n\nThe rule type '{rule}' does not exist. Consider checking your spelling.\nHere's a complete list of valid rules for your reference:\n\n{', '.join(VALID_RULES)}")

        if not all(arg in variations for arg in VALID_RULES[rule]):
            raise ValueError(f'\n\nSeems like you might have forgotten to pass in some arguments for the command rule \'{rule}\'. This requires the following arguments to be passed:\n\n{", ".join(VALID_RULES[rule])}\n\n')

        self.rule_str = rule
        self.kwargs = variations

    def __init_subclass__(cls):
        """
        A custom rule subclass init method.

        Example:
        
        ```py
        class MyRule(CommandRule):
          def handler(self, ctx):
            return True # true or false
    
        MyRule(rule="based.custom") # now it's a valid rule.
        ```
        """
        cls.rule_str = 'based.custom'

    def emit(self, ctx) -> bool:
        return {
            "based.custom": self.handler,
            "cooldown": self.cooldown,
            "except": self._except,
            "for": self._for,
            "usage_limit": self.usage_limit
        }[self.rule_str](ctx)

    def handler(self, ctx) -> bool:
        return True # default, rewritable.
    
    def cooldown(self, ctx) -> bool:
        AUTHOR_ID = ctx.author.id
        
        status: int | None = Tmp.rule_cooldown.get(AUTHOR_ID, None)
        if not status:
            Tmp.rule_cooldown[AUTHOR_ID] = time.time()
            print(Tmp.rule_cooldown)
            return True

        result: bool = (time.time() - status) >= self.kwargs['seconds']

        if result: # if True:
            Tmp.rule_cooldown[AUTHOR_ID] = time.time() # reset
        
        return result

    def _except(self, ctx) -> bool:
        return not ctx.author.id in self.kwargs['users']

    def _for(self, ctx) -> bool:
        return ctx.author.id in self.kwargs['users']

    def usage_limit(self, ctx) -> bool:
        AUTHOR_ID = ctx.author.id
        status = Tmp.rule_usage.get(AUTHOR_ID, None) # times

        if not status:
            Tmp.rule_usage[AUTHOR_ID] = 1

            return True # passed!

        if status > self.kwargs['times']:
            return False # impossible...

        # else: (IGNORE)
        Tmp.rule_usage[AUTHOR_ID] += 1

        return True




class Profile:
    name: str
    id: str
    language: str
    picture_url: str
    status_message: str


class Group:
    count: int
    id: str
    name: str
    picture_url: str

    async def leave():
        """
        Leaves the group. *(New in version 2.2)*
        """



from __future__ import annotations

import re
import uuid
from typing import (List, Literal, Any)
from inspect import iscoroutinefunction as isCoro
from termcolor import colored

from ..exceptions import Async
from ..tmp import Tmp
from ..ext import Depends

# consts
PREFIX = colored('linelib v2', 'light_green')

class ms(int):
  pass

class Decimal(float):
  pass

class TextMessage:
  """
  Represents a text message.

  `text` : str

  The message text (PLAIN TEXT). (e.g., `help me!`)

  ***

  `emojis` : List[Emoji | dict]

  The LINE emojis that are going to be 'injected' into your text.

  To do so, **you will need to make sure to add '$' into your text, just to make us easier to see where we should put a specified emoji.**

  Or, you can use this format: `<product id:emoji id>`. (Designed for lazy people like me)

  **Cool Example 1**

  ```py
  TextMessage(
    "help me... $ I'm stuck in this documentation... $",
    emojis=[
      Emoji('product id', 'emoji id'), # first '$'
      Emoji('product id', 'emoji id') # second '$'
    ]
  )
  ```

  **Epic Example 2**

  ```py
  # mode: lazy
  TextMessage(
    "I am happy. <product id:emoji id> What about you? <product id:emoji id>"
  )
  ```

  > Note: If you really want to disable this kind of mode, try this:

  ```py
  TextMessage(
    "disabled-lazy:Message content here! <wont:work!>"
  )
  ```

  ***

  `quick_reply` : dict | QuickReply

  Quick reply buttons. They will display on the bottom of the screen. 

  ***

  `sender` : dict | Sender

  Send messages with specified profile pictures and names. Note that the name of it will display like this (assuming the original bot name is **Cool Guy**, and the sender name is **Bad Guy**: `'Bad Guy' from Cool Guy` (that's odd)

  """

  def __init__(self,
               text: str,
               emojis: List[Emoji | dict] = None,
               quick_reply: QuickReply = None,
               sender: dict | Sender = None):
    pattern: str = '\<.*\:.*>'
    if re.findall(pattern, text):
      if text.startswith('disabled-lazy:'):
        text = text.replace('disabled-lazy:', '', 1)
      else:
        emojis = []
        for result in re.findall(pattern, text):
          formatted_result = result.replace("<", "").replace(">",
                                                             "").split(":")
          productId = formatted_result[0]  # < Here! : ... >
          emojiId = formatted_result[1]  # < ... : Here! >
          text = text.replace(result, "$")  # cool!
          emojis.append(Emoji(productId, emojiId))
          # ...and then the next part will parse it!

    if emojis:
      parsed_emojis = []
      cur = 0
      for i in re.finditer("\$", text):
        index = i.start()
        json = emojis[cur].json
        json['index'] = index
        parsed_emojis.append(json)
        cur += 1
      # overwrite be like:
      emojis = parsed_emojis  # noqa

    self.json = {
      "type": "text",
      "text": text,
      "emojis": emojis,
      "quickReply": getattr(quick_reply, "json", None) or quick_reply,
      # getattr(sender, "json", None):
      # finds the attribute of "json" of the sender (object?) argument
      # if the attribute was not found, returns None
      # ---------------
      # ... or sender:
      # if the previous statement returns 'None'
      # returns the original sender (JSON)
      "sender": getattr(sender, "json", None) or sender
    }


class Emoji:
  """
  Represents a LINE emoji. You can add it to your `TextMessage` constructors!

  See [List of Available Emojis](https://developers.line.biz/en/docs/messaging-api/emoji-list/)

  `product_id` : str

  The Product ID.

  ***

  `emoji_id` : str

  The Emoji ID.
  
  """

  def __init__(self, product_id: str, emoji_id: str):
    self.json = {
      "productId": product_id,
      "emojiId": emoji_id,
      "index": 'undefined'
    }

  @property
  def product_id(self):
    return self.json['productId']

  @property
  def emoji_id(self):
    return self.json['emojiId']

  @property
  def index(self):
    return self.json['index']


class Sender:
  """
  When sending a message from the LINE Official Account, you can specify the name and the icon_url arguments.

  `name` : str

  Display name.

  ***
  
  `icon_url` : str

  URL of the image to display as an icon when sending a message.
  """

  def __init__(self, name: str = None, icon_url: str = None):
    self.json = {"name": name, "iconUrl": icon_url}


class QuickReply:
  """
  Quicky Quick Reply.

  `items` : List[QuickReplyButton, dict]

  The button objects to display within a row.
  """

  def __init__(self, items: List[QuickReplyButton, dict]):
    done = []
    for i in items:
      done.append(getattr(i, "json", None) or i)
    self.json = {"items": done}


class QuickReplyButton:
  """
  A QuickReply button.

  `action` : PostbackAction | MessageAction | URIAction | DatetimePickerAction | CameraAction | CameraRollAction | LocationAction | RichMenuSwitchAction | dict

  Action performed when this button is tapped. The following is available:

  - Postback
  - Message
  - URI
  - Datetime
  - Camera
  - Camera Roll
  - Location
  - Rich Menu Switch

  `image_url` : str

  URL of the icon at the beginning of the button.
  """

  def __init__(self,
               action: PostbackAction | MessageAction | URIAction
               | DatetimePickerAction | CameraAction | CameraRollAction
               | LocationAction | RichMenuSwitchAction | dict,
               image_url: str = None):
    self.json = {
      "type": "action",
      "action": getattr(action, "json", None) or action,
      "imageUrl": image_url
    }


class HandlableAction:
  def __init__(self, aType):
    self.json = {}
    self.aType = aType
    self.STORE = {}  # : Empty

  async def remember(self, key: Any, item: Any) -> None:
    Tmp.action_storage[self.DATA] = {}
    Tmp.action_storage[self.DATA][key] = item
  
  store = remember

  def handle(self):
    """
    Handles the action.
    """

    def wrapper(func, *args, **kwargs):
      if not isCoro(func):
        raise Async(
          'Async Function',
          f"Function '{func.__name__}' should be an async function.")

      Tmp.self_handler[self.DATA] = func

      return func

    return wrapper

  def store(self, *key, **items):
    """
    Stores further information for this action object. You could get them from whether `@client.event(...)`, or `@ACTION.handle()`! Use `ctx.stored` to fetch them!

    **Example:** `store(epic='yes', super='cool')`

    **Get:** `ctx.stored.super` (returns `"cool"`)
    """
    self.STORE = items  # dict


class PostbackAction(HandlableAction):
  """
  When a control associated with this action is tapped, a postback event is returned via webhook with the specified string in the `data` property.

  `data` : str

  A developer-defined custom ID. You could leave this blank.
  
  `label` : str

  The text displays on the button.

  
  `display_text` : str

  The text that will be sent once the user clicked on the button.

  `input_option`: Literal['close_rich_menu', 'open_rich_menu', 'open_keyboard', 'open_voice']

  The display method of such as rich menu based on user action. Specify one of the following values:

  `close_rich_menu`: Close rich menu
  `open_rich_menu`: Open rich menu
  `open_keyboard`: Open keyboard
  `open_voice`: Open voice message input mode
  
  This is available on LINE version `12.6.0` or later for iOS or Android.

  
  `fill_in_text` : str

  String to be pre-filled in the input field when the keyboard is opened. Valid only when the `input_option` argument is set to `open_keyboard`. The string can be broken by a newline character (`\n`).
  """

  def __init__(self,
               data: str = None,
               label: str = None,
               display_text: str = None,
               input_option: Literal['close_rich_menu', 'open_rich_menu',
                                     'open_keyboard', 'open_voice'] = None,
               fill_in_text: str = None):
    super().__init__('postback')
    if (input_option != "open_keyboard") and fill_in_text:
      raise ValueError(
        "\n\n`input_option` must be 'open_keyboard' in order to make your `fill_in_text` to work."
      )
    self.DATA = data or str(uuid.uuid4())
    self.json = {
      "type": 'postback',
      "data": self.DATA,
      "label": label,
      "displayText": display_text,
      "inputOption": {
        "close_rich_menu": "closeRichMenu",
        "open_rich_menu": "openRichMenu",
        "open_keyboard": "openKeyboard",
        "open_voice": "openVoice"
      }.get(input_option),
      "fillInText": fill_in_text
    }


class MessageAction:
  """
  When a control associated with this action is tapped, the specified text is sent as a message from the user.

  `text` : str

  The text that'll be sent once the button got clicked.

  ***

  `label` : str ~ **Depends**

  Label for the action. The specification depends on which object the action is set to.
  """

  def __init__(self, text: str, label: Depends[str] = None):
    self.json = {"type": "message", "label": label, "text": text}


class URIAction:
  """
  When a control associated with this action is tapped, the URI specified in the uri argument is opened in LINE's in-app browser.

  `uri` : str

  URI opened when the action is performed (Max character limit: 1000). The available schemes are `http`, `https`, `line` *(deprecated)*, and `tel`. **`tel` example: `tel:09001234567`**

  ***

  `label` : str ~ **Depends**

  Label for the action. The specification depends on which object the action is set to.

  ***
  
  `alt_uri_desktop` : str ~ **Depends**

  **IMPORTANT: This is only supported when you set URI actions in Flex Messages, but it doesn't work in quick reply.**
  
  URI opened on LINE for **macOS** and **Windows** when the action is performed (Max character limit: 1000). If the `alt_uri_desktop` argument is set, the `uri` argument is ignored on LINE for macOS and Windows. The available schemes are `http`, `https`, `line` *(deprecated)*, and `tel`. **`tel` example: `tel:09001234567`**
  """

  def __init__(self,
               uri: str,
               label: Depends[str] = None,
               alt_uri_desktop: Depends[str] = None):
    self.json = {
      "type": "uri",
      "label": label,
      "uri": uri,
      "altUri": {
        "desktop": alt_uri_desktop
      }
    }


class DatetimePickerAction(HandlableAction):
  """
  When a control associated with this action is tapped, a postback event is returned via webhook with the date and time selected by the user from the date and time selection dialog. The datetime picker action does not support time zones.

  ***

  ### Date and time format
  The date and time formats for the initial, max, and min values are shown below. The full-date, time-hour, and time-minute formats follow the [RFC3339](https://www.ietf.org/rfc/rfc3339.txt) protocol.

  <table>
    <thead>
      <tr>
        <th>Mode</th>
        <th>Format</th>
        <th>Example</th>
      </tr>
    </thead>
    <tbody>
      <tr>
        <td>date</td>
        <td><code>full-date</code><br />Max: 2100-12-31<br />Min: 1900-01-01</td>
        <td>2017-06-18</td>
      </tr>
      <tr>
        <td>time</td>
        <td><code>time-hour:time-minute</code><br />Max: 23:59<br />Min: 00:00</td>
        <td>12:50</td>
      </tr>
      <tr>
        <td>datetime</td>
        <td><code>full-dateTtime-hour:time-minute or full-datettime-hour:time-minute</code><br />Max: 2100-12-31T23:59<br />Min: 1900-01-01T00:00</td>
        <td>2017-06-18T06:15</td>
      </tr>
    </tbody>
  </table>

  ***

  ### Parameters

  `mode` : Literal['date', 'time', 'datetime']

  Action (picker) mode

  - `date`: Pick date
  - `time`: Pick time
  - `datetime`: Pick date and time

  ***

  `data` : str

  Developer-defined custom ID. You can leave this blank.

  ***

  `label` : str ~ **Depends**

  Label for the action. The specification depends on which object the action is set to.

  ***

  `initial` : str

  Initial (aka. default) value of date or time.

  ***

  `max` : str

  Largest date or time value that can be selected. Must be greater than the `min` value.

  ***

  `min` : str

  Smallest date or time value that can be selected. Must be less than the `max` value.
  """

  def __init__(self,
               mode: Literal['date', 'time', 'datetime'],
               data: str = None,
               label: Depends[str] = None,
               initial: str = None,
               max: str = None,
               min: str = None):
    self.DATA = data or str(uuid.uuid4())
    self.json = {
      "type": "datetimepicker",
      "label": label,
      "data": self.DATA,
      "mode": mode,
      "initial": initial,
      "max": max,
      "min": min
    }


class CameraAction:
  """
  **This action can be configured only with quick reply buttons.** When a button associated with this action is tapped, the camera screen in LINE is opened.
  """

  def __init__(self, label: str):
    self.json = {"type": "camera", "label": label}


class CameraRollAction:
  """
  **This action can be configured only with quick reply buttons.** When a button associated with this action is tapped, the camera roll screen (also known as image selector) in LINE is opened.
  """

  def __init__(self, label: str):
    self.json = {"type": "cameraRoll", "label": label}


class LocationAction:
  """
  **This action can be configured only with quick reply buttons.** When a button associated with this action is tapped, the location screen in LINE is opened.
  """

  def __init__(self, label: str):
    self.json = {'type': 'location', 'label': label}


class RichMenuSwitchAction(HandlableAction):
  """
  **This action can be configured only with rich menus.**

  When you tap a rich menu associated with this action, you can switch between rich menus, and a postback event including the rich menu alias ID selected by the user is returned via a webhook.

  `rich_menu_alias_id` : str

  Rich menu alias ID to switch to.

  ***
  
  `data` : str

  A developer-defined custom ID. You could leave this blank.

  ***
  
  `label` : str ~ Optional

  Action label. Optional for rich menus. Read when the user's device **accessibility feature** is enabled.
  
  """

  def __init__(self, rich_menu_alias_id, data: str = None, label: str = None):
    self.DATA = data or str(uuid.uuid4())
    self.json = {
      "richMenuAliasId": rich_menu_alias_id,
      "data": self.DATA,
      "label": label
    }

class StickerMessage:
  """
  Represents a sticker-only message.

  See [List of Available Stickers ↗](https://developers.line.biz/en/docs/messaging-api/sticker-list/)

  ***

  `package_id` : str

  Package ID for a set of stickers. For information on package IDs, see the List of available stickers.

  `sticker_id` : str

  Sticker ID. For a list of sticker IDs for stickers that can be sent with the Messaging API, see the List of available stickers.
  """
  def __init__(self, package_id: str, sticker_id: str):
    self.json = {
      "type": "sticker",
      "packageId": package_id,
      "stickerId": sticker_id
    }

class ImageMessage:
  """
  Represents an image message.

  Each URL must follow:
  
  - `http` OR `https`
  - `JPEG` or `PNG`
  
  ***

  `original` : str

  The original content URL. (Max file size: `10MB`)

  ***
  
  `preview` : str

  The preview image URL (zipped, for thumbnails). (Max file size: `1MB`)
  """
  def __init__(self, original: str, preview: str):
    self.json = {
      "type": "image",
      "originalContentUrl": original,
      "previewImageUrl": preview
    }

class VideoMessage:
  """
  Represents a video message.

  ***
  
  ### Video Aspect Ratio
  - A very wide or tall video may be cropped when played in some environments.
  - The aspect ratio of the video specified in `original` and the preview image specified in `preview_image` should be the same. If the aspect ratio is different, a preview image will appear behind the video.
  
  ***

  `original` : str

  URL of video file (Max character limit: 2000)
  
  - HTTPS over TLS 1.2 or later
  - mp4
  - Max file size: 200 MB

  ***

  `preview_image` : str

  URL of preview image (Max character limit: 2000)
  
  - HTTPS over TLS 1.2 or later
  - JPEG or PNG
  - Max file size: 1 MB

  ***

  `tracking_id` : str

  ID used to identify the video when Video viewing complete event occurs. If you send a video message with `tracking_id` added, the video viewing complete event occurs when the user finishes watching the video.

  You can use the same ID in multiple messages.

  """
  def __init__(self, original: str, preview_image: str, tracking_id: str = None):
    self.json = {
      "type": "video",
      "originalContentUrl": original,
      "previewImageUrl": preview_image,
      "trackingId": tracking_id
    }

class AudioMessage:
  """
  Represents an audio message.

  ***

  `original` : str

  URL of audio file (Max character limit: 2000)

  - HTTPS over TLS 1.2 or later
  - m4a
  - Max file size: 200 MB

  **Notice**
  ```
  Only M4A files are supported on the Messaging API. If a service only supports MP3 files, you can use a service like FFmpeg to convert the files to M4A.
  ```

  ***

  `duration` : ms

  Length of audio file (milliseconds)
  """
  def __init__(self, original: str, duration: ms):
    self.json = {
      "type": 'audio',
      'originalContentUrl': original,
      'duration': duration
    }

class LocationMessage:
  """
  Represents a location message.

  ***

  `title` : str

  Title, Max character limit: 100. (e.g. `My Location`)

  ***
  
  `address` : str

  Address, Max character limit: 100. (e.g., `1-6-1 Yotsuya, Shinjuku-ku, Tokyo, 160-0004, Japan`)

  ***

  `latitude` : Decimal

  Latitude. (e.g., `35.687574`)

  ***
  
  `longitude` : Decimal

  Longitude. (e.g., `139.72922`)
  
  """
  def __init__(self, title: str, address: str, latitude: Decimal, longitude: Decimal):
    self.json = {
      'type': 'location',
      'title': title,
      'address': address,
      'latitude': latitude,
      'longitude': longitude
    }

class ImagemapMessage:
  """
  [See Documentation (please) ↗](https://developers.line.biz/en/reference/messaging-api/#imagemap-message)
  
  Imagemap messages are messages configured with an image that has multiple tappable areas. You can assign one tappable area for the entire image or different tappable areas on divided areas of the image.

  You can also play a video on the image and display a label with a hyperlink after the video is finished.
  """
  def __init__(self, base_url: str, alt_text: str, base_size: BaseSize | dict, video: Video | dict, actions: List[IMURIAction, IMMessageAction]):
    self.json = { # i hate this >:(
      "baseUrl": base_url,
      "altText": alt_text,
      "baseSize": base_size,
      "video": video,
      "actions": actions
    }


def BaseSize(width: int, height: int):
  return {
    "width": int,
    "height": height
  }


def Video(original: str, preview: str, area: ImagemapArea | dict, external_link: ExternalLink | dict):
  return {
    "originalContentUrl": original,
    "previewImageUrl": preview,
    "area": area,
    "externalLink": external_link
  }

def ImagemapArea(x: int, y: int, width: int, height: int):
  return {
    "x": x,
    "y": y,
    "width": width,
    "height": height
  }

def ExternalLink(link: str, label: str):
  return {
    "linkUri": link,
    "label": label
  }

def IMURIAction(uri: str, area: ImagemapArea | dict, label: str = None):
  return {
    "type": "uri",
    "label": label,
    "linkUri": uri,
    "area": area
  }

def IMMessageAction(text: str, area: ImagemapArea | dict, label: str = None):
  return {
    "type": "message",
    "label": label,
    "text": text,
    "area": area
  }




from __future__ import annotations

class Storage:
  pass

class DatabaseTable:
  """
  A database that helps you to store specific information. You could do this with keys.

  **Example**
  ```py
  # run_me.py
  from linelib.ext import DatabaseTable
  
  DatabaseTable.remember = "I've remembered this!"
  ```
  ```py
  # main.py
  import run_me # runs 'run_me.py'
  from linelib.ext import DatabaseTable

  print(DatabaseTable.remember) # I've remembered this!
  ```
  """
  pass



from __future__ import annotations

import inspect
from typing import Callable

from ..exceptions import MissingArgument, Async, ClientException
from .rule import CommandRule, DEFAULT_RULE, _dfr

class _str:
    def __matmul__(self, other):
        return str

String = _str()

VALID_TYPES = [int, float, bool, str]

class Cog:
    name = "UnnamedCog"
    show_not_found_log = False # whether to show the not found log or not.
    
    def __init__(self):
        self._ll_CONSTRUCTED = True

    def __init_subclass__(cls, prefix: str = None) -> None:
        """
        Initialize a subclass.

        :param str prefix: Command prefix.
        """
        # ll => linelib
        cls._ll_commands: list = []
        cls._ll_CONSTRUCTED: bool = False
        cls.PREFIX: str = prefix or ""
        [cls._ll_commands.append(getattr(cls, i)) if (not i.startswith('__')) and (isinstance(getattr(cls, i), CogCommandWrapper)) else None for i in dir(cls)]

        for item in dir(cls):
          if not item.startswith('__'):
            actual: type = getattr(cls, item) # actual content
            if (
              isinstance(actual, CogCommandWrapper)
              or
              isinstance(actual, CogListenerWrapper)
            ):
              actual.PREFIX = cls.PREFIX
              cls._ll_commands.append(actual)


    async def emit(self, ctx: type):
        gathered = []

        for cmd in self._ll_commands:
            res = await cmd.emit(self, ctx)
            gathered.append(res)

        if all(i == 'no' for i in gathered):
            await self.not_found(ctx, ctx.content.strip().split(' ')[0])
            return 'all-nf' # all not found

    def get_events(self):
      return [
        # handler: str
        # get actual content:
        # getattr(self, handler)
        getattr(self, handler)
        for handler in dir(self)
        if isinstance(getattr(self, handler), CogListenerWrapper)
      ]

    async def emit_events(self, EVENT_TYPE: str, ctx: type):
      for item in self.get_events():
        if item.event == EVENT_TYPE:
          await item.emit(self, ctx)

    async def not_found(self, ctx, command: str):
        """
        Do something when a command was not found.
        """
        pass # default

class CogCommandWrapper:
    """
    Represents a cog command.
    """
    ann: list

    def __init__(self, cmd_name: str, func, rule: CommandRule | _dfr = DEFAULT_RULE, aliases: list[str] = []):
        parameters = inspect.signature(func).parameters

        self.ann = []
        self.rule = rule.emit

        self.aliases = aliases # setup aliases

        keywordOnlyAlreadyFound = False
        cur = 0 # current state

        for name, param in parameters.items():
            cur += 1
            if cur < 3: # meaning that (1, 2) will both be ignored
                continue # ignore

            if (not param.annotation in VALID_TYPES) and (not param.annotation is inspect._empty):
                raise TypeError(f'The type \'{param.annotation.__name__}\' could not be recognized as a valid type.\nLinelib only accepts the following:\nint, float, bool, str')

            if param.kind == inspect.Parameter.KEYWORD_ONLY:
                if keywordOnlyAlreadyFound:
                    raise TypeError('\n\nThe keyword (*) should only be used once in function arguments. This represents that rest of the string that\'s sent from the user will all be passed into this parameter.\n\n')
                else:
                    keywordOnlyAlreadyFound = True
                    # str(name) == parameter name.
                    self.ann.append(('*', str(name)))
            else:
                ann = param.annotation if (not param.annotation is inspect._empty is inspect._empty) else str # empty annotation? then str.

                self.ann.append((ann,))

        self.func: type = func
        self.name = cmd_name
    
    async def emit(self, o: Cog, ctx: type):
        """Emits the command."""

        msg = ctx.content.strip()

        if not msg.startswith(self.PREFIX):
            return "unaccepted-prefix"

        splitted = msg.split(' ')
        
        args = splitted[1:]

        if splitted[0] in [self.name, *self.aliases]:
            should = self.rule(ctx)
            if not should: # not allowed
                await self._RULE_REJECT(o, ctx)
                return
            _PASS = [] # arguments that will be passed in. (*)
            _NAMED = {} # named arguments. (**)

            for i in range(len(args)):
                if (i + 1) > len(self.ann):
                    break # end this
                if self.ann[i][0] != '*':
                    try:
                        _PASS.append(self.ann[i][0](args[i])) # annotations (type)
                    except Exception as err:
                        await self._LL_ERR(o, ctx, err)
                else: # *
                    _NAMED[self.ann[i][1]] = ' '.join(args[i:])
            try:
              await self.func(o, ctx, *_PASS, **_NAMED)
            except Exception as err:
                ERROR = MissingArgument(str(err)) if (isinstance(err, TypeError) and "missing" in str(err)) else err
                await self._LL_ERR(o, ctx, ERROR)

        return 'no'

    def on_error(self, function: Callable) -> Callable:
        if not inspect.iscoroutinefunction(function):
            raise Async("Async Function", f"The function '{function.__name__}' should be an async (coroutine) function. Example:\n\nasync def my_function(...)\n\n")

        self._LL_ERR = function
        return function

    def rule_reject(self, function: Callable) -> Callable:
        if not inspect.iscoroutinefunction(function):
            raise Async("Async Function", f"The function '{function.__name__}' should be an async (coroutine) function. Example:\n\nasync def my_function(...)\n\n")

        self._RULE_REJECT = function
        return function

    async def _RULE_REJECT(self, ctx):
        pass # default

    
    async def _LL_ERR(self, o, ctx, err):
        raise err # default


def cog_command(*, name: String@CogCommandWrapper, rule: CommandRule = DEFAULT_RULE, aliases: list[str] = []) -> CogCommandWrapper:
    """
    Registers a cog command.
    """
    def wrapper(func):
        if not inspect.iscoroutinefunction(func):
            raise Async("Async Function", f"The function '{func.__name__}' should be an async (coroutine) function. Example:\n\nasync def my_function(...)\n\n")
        return CogCommandWrapper(name, func, rule, aliases)

    return wrapper

cog_cmd = cog_command # ~ createAlias

class CogListenerWrapper:
  def __init__(self, event_name: str, func: Callable) -> None:
    self.event = event_name
    self.func = func

  async def emit(self, o: Cog, ctx: type):
    await self.func(o, ctx)
    return 'OK'

def cog_listener(event: str = None) -> Callable:
  if event in ('ready', 'text'):
    raise ClientException('Cogs do not accept the \'ready\' and \'text\' events.\nThe former is unaccepted due to some parameter issues;\nThe latter is due to a conflict would occur when receiving text events & replying to commands.\n\n-> Try modifying the `__init__(self, ...)` function instead.')

  def wrapper(func, *args, **kwargs) -> CogListenerWrapper:
    if not inspect.iscoroutinefunction(func):
      raise Async("Async Function", f"The function '{func.__name__}' should be an async (coroutine) function. Example:\n\nasync def my_function(...)\n\n")
    
    # func.__name__: 'on_text'
    # .split('_'): ['on', 'text']
    # [1]: 'text'
    
    return CogListenerWrapper(event or func.__name__.split('_')[1], func)

  return wrapper

cog_event = cog_listener


"""
Linelib Utils.
"""

from __future__ import annotations

from .model.context import TextMessageEvent, PostbackEvent, StickerMessageEvent, UnsendEvent, FollowEvent, UnfollowEvent, MemberJoinEvent, MemberLeaveEvent, VideoViewingCompleteEvent, BeaconEvent, AccountLinkEvent, ImageMessageEvent, VideoMessageEvent, AudioMessageEvent, FileMessageEvent, LocationMessageEvent, JoinEvent, LeaveEvent

from .tmp import Tmp
from .exceptions import Invalid


class EventObject(object):
  """
  An event object returned from `@client.event`
  """

  def __init__(self, to: dict):
    self.func = to['func']
    self.type = to['type']
    self.options = to['options']  # additional context / options

  async def emit(self, *args, **kwargs):
    await self.func(*args, **kwargs)


def modifiable(*_, **params):
  """
  Represents a function (Any) is modifiable — which means in the subclass, it could be overwritten.
  """

  def wrapper(func):
    return func

  return wrapper


def getTriggerType(req: dict):
  if req['type'] == 'message':
    return req['message']['type']
  else:
    return req['type']


def getContext(_type: str, client, req) -> type:
  context = ({
      "text": TextMessageEvent,
      "postback": PostbackEvent,
      "sticker": StickerMessageEvent,
      "unsend": UnsendEvent,
      "follow": FollowEvent,
      "unfollow": UnfollowEvent,
      "join": JoinEvent,
      "leave": LeaveEvent,
      "memberJoined": MemberJoinEvent,
      "memberLeft": MemberLeaveEvent,
      "videoPlayComplete": VideoViewingCompleteEvent,
      "beacon": BeaconEvent,
      "accountLink": AccountLinkEvent,
      "image": ImageMessageEvent,
      "video": VideoMessageEvent,
      "audio": AudioMessageEvent,
      "file": FileMessageEvent,
      "location": LocationMessageEvent,
  }[_type](
    client,  # Client
    req  # request dictionary (aka. json)
  ))
  context.stored = context.memory = Tmp.handle_action[_type]
  if _type == 'postback':
    context.action_stored = Tmp.action_storage[context.data]

  return context


class utils:
  """
  Linelib utils.
  """

  class _noInit:

    def __init__(self, *args, **kwargs):
      raise Invalid(
        "No Construct",
        "Please do not construct this class. In other words, DO NOT add `(...)` on the right of the class."
      )
    
  class _str:
    def __matmul__(self, other):
      return other
  
  String = _str()

  class QueuedSending(_noInit):
    """
    # Deprecated.
    """
    ACTIVE = True

  class URL(str):
    def __repr__(self):
      return "<linelib URL>"

  class MessageObjects(list):
    """
    Linelib message objects. Should be a list of message classes. (constructed)
    """
    pass


class _method:
  method = "@NotMentioned"

  def __getitem__(self, key):
    self.method = key

  def __repr__(self):
    return f"<Method method='{self.method}'>"


Method = _method()



# Named http.

from __future__ import annotations

from .mkerr import mk
from ..exceptions import ClientException

import httpx # pip install httpx

def url(r: str):
  return "https://api.line.me/v2/" + r.replace('.', '/')

async def reply(client: type, rt: str, msgs: list, disabled: bool):
  """
  rt: reply token
  """
  try:
    async with httpx.AsyncClient() as session:
      r = await session.post(url('bot.message.reply'), headers=client.headers, json={
        "replyToken": rt,
        "messages": msgs,
        "notificationDisabled": disabled
      })
      print(r.json()) # Dev
      resJson = r.json()
      mk(resJson)
      return r
  except Exception as err:
    raise ClientException(err)



from __future__ import annotations

import asyncio  # pip install asyncio
import base64
import hashlib
import hmac
import logging
import random
import time
from inspect import iscoroutinefunction, isfunction
from typing import Any, Callable, Union

from flask import Flask, jsonify, request
from flask_cors import CORS  # pip install flask-cors
from termcolor import colored, cprint

from .construct import (
  EventObject, 
  getContext, 
  getTriggerType, 
  modifiable
)

from .exceptions import Async as AsyncError
from .exceptions import Invalid
from .tmp import Tmp

#from .ext import commands # SHORTCUT-DISALLOWED

# const
prefix = colored('linelib v2', 'light_green')


class Client:
  """
  Represents a Client.
  """

  # public:
  FUNCTION_NAME_OR_LISTENER: str = "@see-func"
  _VALID_EVENTS = ("ready", "text", "postback", "sticker", "unsend", "follow", "unfollow", "join", "leave", "memberJoined", "memberLeft", "videoPlayComplete", "beacon", "accountLink", "image", "video", "audio", "file", "location")
  _EVENTS = {} # Promise

  def __init__(self,
               channel_secret: str,
               channel_access_token: str,
               *args,
               **options: Any):
    self.CS = channel_secret
    self.CAT = channel_access_token

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    self.loop = loop  # if it is required

    for e in self._VALID_EVENTS:
      self._EVENTS[e] = []  # waiting for append
      Tmp.handle_action[e] = {} # database

    self.app = Flask("linelib.Client")
    CORS(self.app)

    self.headers = {
      "Authorization": f"Bearer {self.CAT}",
      "Content-Type": "application/json"
    }  # for quick access, yeah

  def createEvent(self, name: str, obj: type):
    self._EVENTS[name].append(obj)

  def emitEvents(self, name: str, *args: Any, **kwargs: Any):

    async def goOver():
      cache = self._EVENTS[name]
      done = []
      for handler in cache:
        async def doTask():
          if not name in ('ready', ):
            if handler.type == 'postback':
              func = getattr(args[0], 'data', None)
              if func:
                await Tmp.self_handler[func](*args, **kwargs)
                del Tmp.self_handler[func]
          await handler.emit(*args, **kwargs)

          if not name in ('ready', ):
            # args[0] => the context
              
            if getattr(args[0], "TYPE", None) in ['postback', 'datetime', 'rich_menu_switch']:
              try:
                del Tmp.action_storage[args[0].data]
              except Exception as err:
                print(f"\n\n{prefix} - {colored('Unknown Exception ~ TMP_DATA_DELETE', 'red')}:\n{err}\n\n")


          return  # finished, ends the complete task! B)
        
        task = asyncio.create_task(doTask())
        await task

        done.append(task)  # task completed
      res = asyncio.gather(*done)
      return res  # initial response

    loop = self.loop
    task = loop.create_task(goOver())
    loop.run_until_complete(task)

  def event(self,
            listener: Union[str, Callable] = FUNCTION_NAME_OR_LISTENER,
            *options) -> EventObject:
    """
    Represents an event listener. See `Client._VALID_EVENTS` to see a list of valid events.
    """
    if isfunction(listener):
      raise DeprecationWarning(
        "Linelib2 events are now switched to `@client.event()`. Make sure to add the '()' in order to work."
      )

    def wrapper(func, *args, **kwargs):
      if not (iscoroutinefunction(func)):
        raise AsyncError(
          "Async Error",
          f"Function '{func.__name__}(...)' should be an async function.")

      funcName = (listener or func._name__).replace("on_", "")
      if (funcName == "@see-func"):
        funcName = func.__name__.replace("on_", "")

      if (listener in [None, "@see-func"
                       ]) and (not funcName in self._VALID_EVENTS):
        raise Invalid(
          "Invalid Event",
          f"The event name '{funcName}' is not valid.\nConsider checking your spelling, or use the 'print(Client._VALID_EVENTS)' to see all existing event names."
        )

      res = EventObject({
        "func": func,  # handler
        "type": funcName,
        "options": options
      })

      self.createEvent(funcName, res)  # adds event handlers
      # required
      return res

    # required
    return wrapper

  @modifiable(req=dict)  # skip
  def request_then(self, req):
    pass  # default

  @modifiable(ctx=type)
  def payload_then(self, ctx):
    pass  # default

  @modifiable(handler=type, _=Any)
  def emit_before(self, handler, *args, **kwargs):
    """
    Runs before the linelib emits the event handlers.
    """
    pass  # default

  def run(self, **options):
    """
    Runs the LINE bot.

    `**options` - Arguments & options. Note that you must use named options, such as `log_level=ERROR`, `host='0.0.0.0'`, etc.
    """

    @self.app.route("/", methods=['GET', 'POST'])
    def index():
      if request.method == 'GET':
        return "LINE Bot"

      signature = request.headers.get('X-Line-Signature')
      body = request.get_data(as_text=True)
      hash = hmac.new(self.CS.encode('utf-8'), body.encode('utf-8'),
                      hashlib.sha256).digest()
      valid = hmac.compare_digest(signature.encode("utf-8"),
                                  base64.b64encode(hash))
      if not valid:
        raise Invalid(
          "Invalid Signature",
          f"\n\nAn invalid request was received. INFO:\nUser-Agent: {request.headers.get('User-Agent')}\nSignature Received: {signature}\nBody:\n{body}"
        )

      req: dict = request.json

      self.request_then(req)

      # requested
      for payload in req['events']:
        #print(payload) # Dev
        t = getTriggerType(payload)
        context: type = getContext(
          t,
          self,
          payload  # duplicate
        )  # update: NOT *getContext(...)
        self.payload_then(context)
        self.emitEvents(t, context)

      if req['events'] == []:  # empty array
        cprint("\n\n🎉 Surprise!", "light_green")
        cprint(
          "Congratulations on your success — your webhook URL has been verified by LINE.\n\n"
        )
        return jsonify({"status": "Okay"})

      return jsonify({"status": "OK", "message": "Request accepted (linelib)"})

    START = time.time()

    @self.app.before_first_request
    def before_first():
      OKAY = time.time()
      TAKEN = OKAY - START
      if options.get('show_logs') in [None, True]:
        facts = [
          "LINE API has a slow delivery.", "Congrats on your success!",
          "I like chocolate.", "The 'New' method is deprecated.",
          "Try out LINE Notify with linelib!",
          "It's easy to view stats of your bot."
        ]
        print(
          f"\n\n{prefix} - App Running!\n{' ' * len('linelib v2 - ')}{random.choice(facts)}\n\n✨ {colored('ready', 'blue')} in {round(TAKEN * 1000)} ms\n\n"
        )

      self.emitEvents("ready")

    # launch application
    if options.get('show_logs') in [None, True]:
      print(
        f"\n\n{prefix} - Waiting for request...\n{' ' * len('linelib v2 - ')}Consider {colored('refreshing the webview.', 'red')}\n\n"
      )

    # option validation
    opt = options.copy()

    logging.getLogger('werkzeug').setLevel(logging.ERROR if (
      not "log_level" in options) else options.get('log_level'))

    if opt.get('log_level'):
      del opt['log_level']

    START = time.time()  # re-assign
    self.app.run(**opt, host="0.0.0.0", port=8080)

  def load_cog(self, cog: type):
    if not cog._ll_CONSTRUCTED:
      raise Invalid('Invalid Command Cog', f'We found out that your cog \'{cog.__name__}\' was not constrcted.\nPlease use `client.load_cog({cog.__name__}())` instead.')

    @self.event('text')
    async def on_mount(ctx):
        res = await cog.emit(ctx)
        if res == 'all-nf' and cog.show_not_found_log:
            print('\n' + colored('linelib v2', 'light_green') + " " + colored('COG', attrs=['bold', 'underline']) + f" - {cog.name}\n{' ' * len('linelib v2 ')}" + colored(f'Command \'{ctx.content.strip().split(" ")[0]}\' not found.', 'red') + '\n')

    for event in cog.get_events():
      print(event)
      @self.event(event.event)
      async def on_event(ctx):
        await event.emit(cog, ctx)
    return ...

  load_extension = load_cog

  # ================================

  async def sleep(self, seconds: int | float) -> None:
    """
    An asyncio method of `time.sleep`
    """
    await asyncio.sleep(seconds)



# hello if ur on github


# make error description
from ..exceptions import ClientException


def mk(resJson: dict):
    if "message" in resJson: # some errors in there
        errs = ""
        if not 'details' in resJson:
          raise ClientException(f"""Error while making a request:\n ~ {resJson['message']}""")
        for e in resJson['details']:
            errs += f"• {e['message']}\n~~~~AT `{e['property']}`\n\n"

        raise ClientException(
             f"""\n\nError while making a request:\n ~ {resJson['message']}\n\n{errs}"""
        )



from __future__ import annotations

import asyncio
import time
from termcolor import colored
from typing import Any, Literal

from .message import WBTextMessage, WBPostback, WBSticker

from ..connect.gate import reply as replyFunc
from ..connect.fetch import profile, profileAndGroup, getContent
from ..connect.types import Group, Profile

from ..exceptions import Usage as UsageError

from .new import TextMessage
from ..tmp import Tmp
from ..ext import Depends

class MessageObjects(list):
  """
  Message Objects based on a list.
  """

class BaseEvent:
  """
  Represents an event parent class.
  """
  # public:
  reply_token: str
  is_redelivery: bool
  event_id: str
  timestamp: int
  ping: float
  queued: bool = False
  author: Profile
  room_type: str
  
  def __init__(self, CAT: str, req: dict) -> BaseEvent:
    self.CAT = CAT
    self.req = req

    self.reply_token = req.get('replyToken', None)
    if not self.reply_token:
        self.reply = self.send = None
    self.is_redelivery: bool = req['deliveryContext']['isRedelivery']
    self.event_id: str = req['webhookEventId']
    self.timestamp: int = req['timestamp']
    self.ping = ((time.time() * 1000) - req['timestamp'])
    self.replied = False
    self.t_queue = []


    async def init():
        if req['source']['type'] == 'user':
            self.room_type = 'user'
            
            self.author = self.user = await profile({
                  "Authorization": "Bearer " + CAT
            }, req['source']['userId'])
            self.get_group = self.fetch_group = None
        else:
            self.room_type = 'group' # or multi-persons chat
            _profile, _groupCoroutineFunc = await profileAndGroup({
                "Authorization": "Bearer " + CAT
            }, req['source']['userId'], req['source']['groupId'])
            self.get_group = self.fetch_group = _groupCoroutineFunc
            self.author = self.user = _profile

    asyncio.run(init())

  async def get_group(self) -> Group | None:
      """
      (*New in version 2.2*)

      Fetches the current group. If this is a normal chatroom, this function does not exist.
      """
      pass # default

  async def reply(self, messages: MessageObjects = None, notification_disabled: bool = False, *args, **kwargs):
    """
    Replies the user.

    `messages` : MessageObjects

    The messages you want to send. `list` expected.

    ***

    `notification_disabled` : bool

    Whether to disable the push notifications for users. To disable, use `True`. Note that if the user has muted the chat / group, they will not receive any messages even though this is set to `False`.

    `*args, **kwargs` Arguments.
    """
    if not messages:
      messages = [
        TextMessage(*args, **kwargs)
      ]
    if not isinstance(messages, list):
      messages = [messages] # stupid way but why not

    if self.replied:
      raise UsageError(
        "Usage Error", f"You could only use the reply function once. To send multiple messages, try this:\n\n{colored('await', 'magenta')} ctx.{colored('send', 'yellow')}([\n  message1,\n  message2\n])\n\n"
      )

    for item in range(len(messages)):
      if isinstance(messages[item], str):
        messages[item] = TextMessage(messages[item]).json
        # in case we have to waste our time.

      if not isinstance(messages[item], dict):
        messages[item] = messages[item].json
        # the waffle house has found its new host

    await replyFunc(self.client, self.reply_token, messages, notification_disabled)

  async def remember(self, key: Any, item: Any):
    Tmp.handle_action[self.TYPE][key] = item

  store = remember
  send = reply


class TextMessageEvent(BaseEvent):
  """
  A LINE Text Message context.
  """
  message: WBTextMessage
  content: str
  text: str
  id: str
  client: type
  TYPE = 'text'

  def __init__(self, client, r):
    super().__init__(client.CAT, r)
    self.replied = False
    self.t_queue = []
    self.message = WBTextMessage(self.req)
    self.content = self.text = self.message.content # createAlias
    self.id = self.message.id
    self.client = client

class PostbackEvent(BaseEvent):
  """
  A LINE Postback Event.
  """
  data: str
  datetime: str | None
  rich_menu: str | None
  TYPE = 'postback'
  
  def __init__(self, client, r):
    super().__init__(client.CAT, r)
    self.client = client
    self.postback = WBPostback(r)
    self.data: str = self.postback.data
    self.datetime: str | None = self.postback.datetime
    self.rich_menu: str | None = self.postback.rich_menu

class StickerMessageEvent(BaseEvent):
    """
    A LINE Sticker Message Event.
    """
    sticker: WBSticker
    sticker_id: str
    package_id: str
    keywords: list[str]
    resource_type: Literal['static', 'animation', 'sound', 'animation_sound', 'popup', 'popup_sound', 'custom', 'message']
    TYPE = 'sticker'
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        self.client = client
        sticker = WBSticker(r)
        
        self.sticker_id = sticker
        self.package_id = sticker.package_id
        
        self.keywords = sticker.keywords
        self.resource_type = sticker.resource_type

        self.id = r['message']['id']

class UnsendEvent(BaseEvent):
    """
    Represents an unsend event.
    """
    message_id: str
    TYPE = 'unsend'
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        self.client = client
        self.message_id = self.id = r['unsend']['messageId']

class FollowEvent(BaseEvent):
    """
    Event object for when your LINE Official Account is added as a friend (or unblocked). You can reply to follow events.
    """
    TYPE = 'follow'
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        
        self.client = client
        
class UnfollowEvent(BaseEvent):
    """
    Event object for when your LINE Official Account is blocked.

    Note that **you cannot reply to this kind of event**.
    """
    TYPE = 'unfollow'
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)

        class Unavailable:
            def __str__(self):
                return '...'

            def __repr__(self):
                return '...'
        
        self.client = client

class JoinEvent(BaseEvent):
    """
    Event object for when your LINE Official Account joins a group chat or multi-person chat. You can reply to join events.

    - For group chats: A join event is sent when a user invites your LINE Official Account.

    - For multi-person chats: A join event is sent when the first event (for example when a user sends a message or is added to the multi-person chat) occurs after your LINE Official Account is added.
    """
    TYPE = "join"
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)

        self.client = client

class LeaveEvent(BaseEvent):
    """
    Event object for when a user removes your LINE Official Account from a group chat or when your LINE Official Account leaves a group chat or multi-person chat.
    """
    TYPE = "leave"
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)

        self.client = client

class MemberJoinEvent(BaseEvent):
    """
    Event object for when a user joins a group chat or multi-person chat that the LINE Official Account is in.
    """
    TYPE = "memberJoined"
    members: list

    def __init__(self, client, r):
        super().__init__(client.CAT, r)

        members = r['joined']['members']
        self.members = self.joined_members = [user['userId'] for user in members]

        self.client = client

class MemberLeaveEvent(BaseEvent):
    """
    Event object for when a user leaves a group chat or multi-person chat that the LINE Official Account is in.
    """
    TYPE = "memberLeft"

    def __init__(self, client, r):
        super().__init__(client.CAT, r)

        members = r['left']['members']
        self.members = self.left_members = [user['userId'] for user in members]

        self.client = client

class VideoViewingCompleteEvent(BaseEvent):
    """
    Event for when a user finishes viewing a video at least once with the specified `tracking_id` sent by the LINE Official Account.

    ## The number of video views
    A video viewing complete event doesn't necessarily indicate the number of times a user has watched a video.

    Watching a video multiple times in a single session in a chat room doesn't result in a duplicate event. However, if you close the chat room and reopen it to watch the video again, the event may reoccur.

    ## Video in imagemap messages and flex messages is not supported by the video viewing complete event
    The `tracking_id` can't be specified for a video in imagemap messages and flex messages.
    """
    TYPE = "videoPlayComplete"
    tracking_id: str

    def __init__(self, client, r):
        super().__init_(client.CAT, r)
        self.tracking_id = r['videoPlayComplete']['trackingId']

        self.client = client

class BeaconEvent(BaseEvent):
    """
    Event object for when a user enters the range of a [LINE Beacon](https://developers.line.biz/en/docs/messaging-api/using-beacons/). You can reply to beacon events.

    ## Possible Beacon Events
    - `enter` - Entered beacon's reception range.
    - `banner` - Tapped beacon banner. (Cooperated with LINE only)
    - `stay` - A user is within the range of the beacon's reception. This event is sent repeatedly at a minimum interval of 10 seconds.
    """
    TYPE = "beacon"
    hwid: str
    beacon_type: str
    dm: str | None

    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        self.hwid = self.hardware_id = r['beacon']['hwid']
        self.type = self.beacon_type = r['beacon']['type']
        self.dm = self.device_message = r['beacon'].get('dm', None)
        self.client = client

class AccountLinkEvent(BaseEvent):
    """
    Event object for when a user has linked their LINE account with a provider's service account. You can reply to account link events.
    """
    TYPE = "accountLink"
    result: Literal['ok', 'failed']
    nonce: str
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        if not self.reply_token:
            # This property won't be included if linking the account has failed
            self.reply = self.send = None
        
        self.result: Literal['ok', 'failed'] = r['link']['result']
        self.nonce: str = r['link']['nonce']

        self.client = client

class SavableFile(BaseEvent):
    """
    This file could be downloaded and saved in your (local) environment.

    *New in v2.2*
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def save(self, /) -> str:
        """
        Save the document (file).

        Returns: file name.
        """
        # fn: file name
        fn = await getContent(self.client.headers, self.id)
        return fn

class ImageMessageEvent(SavableFile):
    """
    Message object which contains the image content sent from the source.
    """
    TYPE = 'image'
    id: str
    content_provider: Literal['line', 'external']
    original_content_url: Depends[str]
    preview_image_url: Depends[str]
    image_sets: Depends[list]

    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        req: dict = r['message']
        self.req = req
        self.client = client

        self.id = req['id']
        self.image_sets = req.get('imageSet', None)

        self.content_provider = self.provider = req['contentProvider']['type']
        
        if self.content_provider == 'external':
            self.original_content_url = req['contentProvider']['originalContentUrl']
            self.preview_image_url = req['contentProvider']['previewImageUrl']

class VideoMessageEvent(SavableFile):
    """
    Message object which contains the video content sent from the source. The preview image is displayed in the chat and the video is played when the image is tapped.
    """
    TYPE = 'video'
    id: str
    content_provider: Literal['line', 'external']
    original_content_url: Depends[str]
    preview_image_url: Depends[str]
    duration: int # mileseconds

    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        req: dict = r['message']
        self.req = req
        self.client = client

        self.id = req['id']
        self.duration = req.get('duration', None)
        self.content_provider = self.provider = req['contentProvider']['type']
        
        if self.content_provider == 'external':
            self.original_content_url = req['contentProvider']['originalContentUrl']
            self.preview_image_url = req['contentProvider']['previewImageUrl']

class AudioMessageEvent(SavableFile):
    """
    Message object which contains the audio content sent from the source.
    """
    TYPE = 'audio'
    id: str
    content_provider: Literal['line', 'external']
    original_content_url: Depends[str]
    preview_image_url: str
    duration: int # mileseconds

    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        req: dict = r['message']
        self.req = req
        self.client = client

        self.id = req['id']
        self.duration = req.get('duration', None)
        self.content_provider = self.provider = req['contentProvider']['type']
        if self.provider == 'external':
            self.original_content_url = req['contentProvider']['originalContentUrl']



class FileMessageEvent(SavableFile):
    """
    Message object which contains the file sent from the source. The binary data of the file can be retrieved by specifying the message ID and calling the API.
    """
    TYPE = 'file'
    file_name: str
    file_size: int # in bytes
    
    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        req: dict = r['message']
        self.req = req

        self.id = req['id']
        self.file_name = self.name = req['fileName']
        self.file_size = self.size = req['fileSize']

        self.client = client

class LocationMessageEvent(BaseEvent):
    """
    Message object which contains the location data sent from the source.
    """
    TYPE = 'location'
    id: str
    title: Depends[str]
    address: Depends[str]

    latitude: float
    longitude: float

    def __init__(self, client, r):
        super().__init__(client.CAT, r)
        self.req = req = r['message']
        self.id = req['id']
        
        self.title = req.get('title', None)
        self.address = req.get('address', None)
        
        self.latitude = req['latitude']
        self.longitude = req['longitude']

        self.client = client



# hello again!
# im really bored


"""
LineLib version 2.

https://github.com/AWeirdScratcher/linelib
"""

from .client import *
from .exceptions import *
from .construct import *
from .model.context import *
from .model.new import *
from .model.message import *

__author__ = "AWeirdScratcher (AWeirdDev)"
__version__ = "2.0"


# throw

class Base(Exception):
  def __init__(self, title, txt):
    txt = f"\n\n[{title}]\n" + txt
    super().__init__(txt) # end
 
class Async(Base):
  """
  Async related exceptions. (mainly)
  """
  pass

class Invalid(Base):
  """
  Anything that's invalid.
  """
  pass

class Usage(Base):
  """
  Usage exceptions.
  """
  pass

class ClientException(Exception):
  """
  A client-based exception. (Things that are related to the Client.)
  """
  pass

class MissingArgument(Exception):
    """
    Represents a command argument was not passed into.
    """
    pass


"""
fetch fetch fetch fetch
"""

from __future__ import annotations
import httpx # pip install httpx
import termcolor
import uuid
import os
from tqdm import tqdm
from mimetypes import guess_extension

from .mkerr import mk
from ..exceptions import ClientException

def fetch_this(thing: str):
  return "https://api.line.me/v2/bot/" + thing.replace('.', '/')

async def profile(headers: dict, user_id: str) -> type:
  async with httpx.AsyncClient() as session:
    res = await session.get(fetch_this(f'profile.{user_id}'), headers=headers) # response
    json = res.json() # the response json
    mk(json)
      
    class Profile:
      name = display_name = json['displayName']
      id = user_id = json['userId']
      language = region = json['language']
      picture_url = picture = avatar_url = avatar = json['pictureUrl']
      status_message = status = json['statusMessage']
      
    return Profile() # without brackets...? well, it works too.

async def profileAndGroup(headers: dict, user_id: str, group_id: str) -> type:
    async with httpx.AsyncClient() as session:
        res = await session.get(fetch_this(f'profile.{user_id}'), headers=headers) # response
        json = res.json() # the response json
        mk(json)
        
        class Profile:
          name = display_name = json['displayName']
          id = user_id = json['userId']
          language = region = json['language']
          picture_url = picture = avatar_url = avatar = json['pictureUrl']
          status_message = status = json['statusMessage']

        async def get_group():
            async with httpx.AsyncClient() as session:
                GS_RES = await session.get(fetch_this(f'group.{group_id}.summary'), headers=headers)
                groupSum = GS_RES.json() # group summary (json)

                mk(groupSum)

                MC_RES = await session.get(fetch_this(f'group.{group_id}.members.count'), headers=headers)
                groupMemberCount = MC_RES.json()['count']

                class Group:
                    count: int = groupMemberCount
                    id: str = groupSum['groupId']
                    name: str = groupSum['groupName']
                    picture_url = picture = groupSum['pictureUrl']

                    async def leave(self):
                        return await leave_gr(headers, group_id)

                return Group()

        return (Profile(), get_group)

async def leave_gr(headers: dict, group_id: str):
    """
    Leave a `gr`oup chat.
    """
    try:
        async with httpx.AsyncClient() as session:
            res = await session.post(fetch_this(f'group.{group_id}.leave'), headers=headers)
            mk(res.json())
            return res
    except Exception as err:
        raise ClientException(err)


async def getContent(headers: dict, message_id: str):
    """
    Fetches the message content. (images, videos, audio, and files)
    """
    fileName = "linelib-dl-" + str(uuid.uuid4()).split('-')[0] + ".TMP"
    try:
        client = httpx.AsyncClient()
        with open(fileName, 'wb') as file: # FILE !important
            async with client.stream('GET', f"https://api-data.line.me/v2/bot/message/{message_id}/content", headers=headers) as response:
                print("\n\n" + termcolor.colored('linelib v2', 'light_green') + "- 📦 Downloading files...")
                
                fileExtension = guess_extension(response.headers['content-type'].partition(';')[0].strip())
                total = int(response.headers["Content-Length"])
                with tqdm(total=total, unit_scale=True, unit_divisor=1024, unit="file") as progress:
                    num_bytes_downloaded = response.num_bytes_downloaded
                    async for chunk in response.aiter_bytes():
                        file.write(chunk)
                        progress.update(response.num_bytes_downloaded - num_bytes_downloaded)
                        num_bytes_downloaded = response.num_bytes_downloaded

                print(termcolor.colored('linelib v2', 'light_green') + ' - Finished!\n') # finished

                file.close()
                # .TMP => length: 4
                newName: str = fileName[:-4] + "." + fileExtension
                os.rename(fileName, newName)
                return newName
    
    except Exception as err:
        raise ClientException(err)


from __future__ import annotations

import aiohttp

from ..construct import utils

def _api(url):
  return "https://notify-api.line.me/api/" + url.replace('.', '/')

class Notify:
  """
  [LINE Notify.](https://notify-bot.line.me)
  """
  def __init__(self, access_token: str):
    self.AT: str = access_token
    self.headers: dict = {
      "Authorization": "Bearer " + access_token,
      "Content-Type": "application/x-www-form-urlencoded"
    }

  async def notify(self, message: str, image_thumbnail: utils.URL = "", image_full_size: utils.URL = "", notification_disabled: bool = False):
    """
    Send a notification through LINE Notify.

    message : str

    The message content. (MAX: 1000 characters)

    image_thumbnail, image_full_size : URL

    Should be both set at the same time.

    - `image_thumbnail`: 240×240px JPEG
    - `image_full_size`: 2048×2048px JPEG

    notification_disabled : bool

    When set to `True`, the user will not receive a push message. However, even though it's set to `False` -- which means normally you will receive a push message -- if some users muted the group / bot, this will not work.
    """
    if image_thumbnail:
      image_full_size = image_thumbnail
    elif image_full_size:
      image_thumbnail = image_full_size


    async with aiohttp.ClientSession() as session:
      async with session.post(_api('notify'), headers=self.headers, params={
        "message": message,
        "imageThumbnail": image_thumbnail,
        "imageFullsize": image_full_size,
        "notification_disabled": {True: "true", False: "false"}[notification_disabled]
      }) as resp:
        return resp

  async def status(self):
    async with aiohttp.ClientSession() as session:
      async with session.get(_api('status'), headers=self.headers) as resp:
        json = await resp.json()
        
        class Response:
          status: int = resp.status
          message: str = json['message']
          target_type: str = json['targetType']
          target: str = json['target']

        return Response


from distutils.core import setup

global README

with open('README.md', 'r', encoding="utf-8") as f:
  README = f.read()

setup(
    name='linelib',
    version='2.2.4',
    packages=['linelib', 'linelib.notify', 'linelib.ext', 'linelib.connect', 'linelib.model'],
    license='MIT',
    description="The solution to simplicity.",
    long_description=README,
    long_description_content_type="text/markdown",
    author='AWeirdScratcher',
    author_email = "aweirdscrather@gmail.com",
    install_requires=[
        'httpx', 'urllib3', 'flask', 'flask-cors', 'termcolor', 'tqdm'
    ],
    keywords = ['line', 'bot', 'line bot', 'sdk', 'line notify', 'commands'],
    url="https://github.com/AWeirdScratcher/linelib",
    classifiers=[
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.10'
  ],
    entry_points={
      'console_scripts': [
        'linelib=linelib.__main__:main'
      ]
    }
)


"""
Temporary database. This functionality was named `database` in the earlier version.
"""

class Tmp:
    """
    Temporary Database / Storage.
    """
    handle_action = {}
    action_storage = {} # :Empty
    self_handler = {} # :Empty
    
    rule_cooldown = {}
    rule_usage = {}


"""
Create your first linelib project through your terminal.
"""
from __future__ import annotations

import click # pip install click
import requests # pip install requests
import uuid
import os
import time
import sys

from flask import Flask

"""
Configure here!
"""
AVAILABLE_EXAMPLES = {
    'hello': [
        'https://gist.githubusercontent.com/AWeirdScratcher/8ddac68554328b8beaa139d8c4f01e79/raw/ca669b3b37ee4952ea370270de802ff08025957d/main.py'
    ]
}
FILE_NAMES = {
    'hello': ['main.py']
}



@click.group()
def cli():
    pass


@cli.command()
def version():
    """Shows version information."""
    click.secho('\n⚙️  linelib v2 CLI', fg="green")
    click.echo(f'{sys.version}\n\n• CLI Version: 0.1\n• Run \'py linelib\' for command help.\n')
    click.echo('- Created by AWeirdScratcher (AWeirdDev)\n')

@cli.command()
@click.option('-project', '--project', '-p', '--p', required=True)
def create(project: str):
    """
    Create a directory with Linelib examples.
    """
    if not project in AVAILABLE_EXAMPLES:
        return click.secho(f'\n\nThe example project name \'{project}\' is not available, or has been removed.\n\n', fg="red")
    click.echo('\n\nCloning example \'' + project + '\'...\n\n  • Session started.')

    fetched: list = []

    start: float = time.time()
    try:
        for url in AVAILABLE_EXAMPLES[project]:
            r = requests.get(url)
            fetched.append(r.text)
    except Exception as err:
        sN = '\n'
        click.secho('  \nFailed while fetching:', fg="yellow")
        click.echo(f'  {str(err).replace(sN, sN + "  ")}')
        click.secho('\nFailed to fetch the project. See above for more information.\n\n', fg="red")
        return False
    
    FOLDER: str = "linelib-" + str(uuid.uuid4()).split('-')[0]

    click.echo('  • Fetch complete.')
    click.echo(f'  • Appending files... ({len(FILE_NAMES[project])})')
    os.mkdir(FOLDER)

    current = 0

    for file in FILE_NAMES[project]:
        with open(FOLDER + "/" + file, "a") as f:
            f.write(fetched[current])
            f.close()
        current += 1

    end: float = time.time()
    click.echo(f'\n  → Execution Completed! (Time used: {round((end - start) * 1000)}ms)')
    click.secho(f'\n\n  Folder name: {FOLDER}', fg="blue")
    click.secho('\nDone. (100%)\n\n', fg="green")

@cli.command()
def start():
    """
    Start the development server.
    """
    click.secho('\n\nlinelib - v2', fg="green")
    click.echo('  • Starting development server...\n')

    app = Flask(__name__)

    @app.route('/', defaults={'path': ''})
    @app.route('/<path:path>')
    def web_application(path):
        return "<center>\n<h1 style=\"font-family: sans-serif; font-size: 3rem;\">Hello, World!</h1>\n<p>Specified path: /" + path + "</p>\n</center>"
    
    @app.before_first_request
    def first_req():
        click.secho('\n  • Caught: first request.\n', fg="blue")

    app.run(host="0.0.0.0", port=8080)


def main():
  if __name__ == "__main__":
      cli()